# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest.mock import Mock

from qtpy.QtWidgets import QWidget
from traits.api import Instance

from karabo.common.scenemodel.bases import BaseEditWidget, BaseWidgetObjectData
from karabo.native import Configurable, Hash, Int8, Int32
from karabogui.binding.api import Int32Binding
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller)
from karabogui.sceneview.widget.api import ControllerContainer
from karabogui.testing import get_property_proxy, singletons


class IntDevice(Configurable):
    base = Int32(
        defaultValue=0)
    rotary = Int32(
        defaultValue=0)
    linear = Int32(
        defaultValue=0)
    forbidden = Int8(
        defaultValue=0)


class DisplayModel(BaseWidgetObjectData):
    """A test display model for controllers"""


class EditableModel(BaseEditWidget):
    """A test editable model for controllers"""


@register_binding_controller(can_edit=False, klassname="TestDisplay",
                             binding_type=Int32Binding)
class DisplayController(BaseBindingController):
    model = Instance(DisplayModel, args=())

    def create_widget(self, parent):
        return QWidget(parent)

    def add_proxy(self, proxy):
        return True

    def remove_proxy(self, proxy):
        return False


@register_binding_controller(can_edit=False, klassname="TestEditable",
                             binding_type=Int32Binding)
class EditableController(BaseBindingController):
    model = Instance(EditableModel, args=())

    def create_widget(self, parent):
        return QWidget(parent)

    def add_proxy(self, proxy):
        return True

    def remove_proxy(self, proxy):
        return True


def test_display_widget_container(gui_app):
    """Test the controller container with display controller"""
    model = DisplayModel(keys=["TestDevice.base", "TestDevice.rotary"])
    container = ControllerContainer(DisplayController, model, None)

    controller = container.widget_controller
    assert controller.widget is not None
    assert container.toolTip() == "TestDevice.base, TestDevice.rotary"
    proxy = get_property_proxy(IntDevice.getClassSchema(), "linear")

    assert container.add_proxies([proxy])
    assert len(controller.proxies) == 3
    tooltip = "TestDevice.base, TestDevice.rotary, TestDevice.linear"
    assert container.toolTip() == tooltip

    # We are a readonly container, nothing gets send outside
    network = Mock()
    manager = Mock()
    with singletons(network=network, manager=manager):
        container.apply_changes()
        network.onReconfigure.assert_not_called()
        # Apply when we are visible
        container.set_visible(True)
        container.apply_changes()
        network.onReconfigure.assert_not_called()
        proxy.edit_value = 2
        container.apply_changes()
        network.onReconfigure.assert_not_called()

    # Try to add a forbidden proxy
    forbidden = get_property_proxy(IntDevice.getClassSchema(), "forbidden")
    assert not container.add_proxies([forbidden])
    assert len(controller.proxies) == 3
    assert container.toolTip() == tooltip

    # Try to remove although controller does not allow
    assert not container.remove_proxies([proxy])
    assert len(controller.proxies) == 3
    assert container.toolTip() == tooltip
    container.destroy()


def test_editable_widget_container(gui_app):
    """Test the controller container with editable controller"""
    model = EditableModel(keys=["TestDevice.base", "TestDevice.rotary"])
    container = ControllerContainer(EditableController, model, None)

    controller = container.widget_controller
    assert controller.widget is not None

    assert container.toolTip() == "TestDevice.base, TestDevice.rotary"
    proxy = get_property_proxy(IntDevice.getClassSchema(), "linear")
    assert container.add_proxies([proxy])
    assert len(controller.proxies) == 3
    assert container.toolTip() == (
        "AccessLevel: OPERATOR - Access: True\n\nTestDevice.base, "
        "TestDevice.rotary, TestDevice.linear")
    network = Mock()
    manager = Mock()
    with singletons(network=network, manager=manager):
        # Reconfigure single proxy
        proxy.edit_value = 2
        container.apply_changes()
        network.onReconfigure.assert_not_called()
        container.set_visible(True)
        container.apply_changes()
        network.onReconfigure.assert_called_with("TestDevice",
                                                 Hash("linear", 2))
        network.reset_mock()
        # Reconfigure twice
        container.apply_changes()
        network.onReconfigure.assert_called_with("TestDevice",
                                                 Hash("linear", 2))
        network.reset_mock()
        # Decline changes
        assert proxy.edit_value is not None
        container.decline_changes()
        assert proxy.edit_value is None
        container.apply_changes()
        network.onReconfigure.assert_not_called()

    assert container.remove_proxies([proxy])
    assert len(controller.proxies) == 2
    container.destroy()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import json
from os import path as op

from karabogui.dialogs import update_dialog
from karabogui.dialogs.update_dialog import UNDEFINED_VERSION, UpdateDialog

_PKG = "GUIExtensions"


class MockReturn:
    def __init__(self, info: dict):
        self._info = info

    def json(self) -> dict:
        return self._info


def update_pkg(package_name: str):
    if package_name == _PKG:
        return "1.0.0"
    return UNDEFINED_VERSION


def test_update_dialog(gui_app, mocker):
    """Tests the model of the extension updater"""
    mocker.patch.object(update_dialog, "get_pkg_version",
                        side_effect=update_pkg)
    request = mocker.patch("requests.get")
    request.return_value = MockReturn({})
    dialog = UpdateDialog()
    assert dialog.label_current.text() == "1.0.0"
    assert dialog.label_latest.text() == UNDEFINED_VERSION


def test_dialog_index_parsing(gui_app, mocker):
    """Test the proper parsing of a json file from the index"""
    json_file = op.join(op.dirname(__file__), "test_json.json")
    with open(json_file) as f:
        package = json.loads(f.read())

    mocker.patch("requests.get", return_value=MockReturn(package))
    packages = update_dialog.get_index_list()
    assert packages.get(_PKG) == "0.0.1"



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from platform import system

import pytest
from qtpy.QtWidgets import QLayout

from karabo.common.scenemodel.api import DisplayTextLogModel
from karabo.native import Configurable, String
from karabogui.events import KaraboEvent
from karabogui.testing import get_class_property_proxy, set_proxy_value

from ..textlog import DisplayTextLog


class Object(Configurable):
    prop = String()


def get_widgets_from_layout(layout: QLayout) -> list:
    widgets = []
    for i in range(layout.count()):
        item = layout.itemAt(i)

        widget = item.widget()
        if widget is not None:
            widgets.append(widget)

        child_layout = item.layout()
        if child_layout is not None:
            widgets.extend(get_widgets_from_layout(child_layout))

    return widgets


@pytest.mark.skipif(system() == "Windows",
                    reason="toPlainText returns an empty string in windows")
def test_textlog_controller(gui_app, mocker):
    # setup
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    controller = DisplayTextLog(proxy=proxy, model=DisplayTextLogModel())
    controller.create(None)
    assert controller.widget is not None
    # set value
    controller.log_widget.clear()
    set_proxy_value(proxy, "prop", "Line 1")
    assert "Line 1" in controller.log_widget.toPlainText()
    widgets = get_widgets_from_layout(controller.widget.layout())
    assert widgets[0].toolTip() == ""
    assert widgets[1].toolTip() == "Historic Text Log"
    broadcast = mocker.patch(
        "karabogui.controllers.display.textlog.broadcast_event")
    widgets[1].click()
    broadcast.assert_called_once()
    args, kwargs = broadcast.call_args
    event = args[0]
    assert event is KaraboEvent.ShowUnattachedController
    assert "model" in args[1]
    assert args[1]["model"].keys == ["prop"]
    assert len(widgets) == 3
    # teardown
    controller.destroy()
    assert controller.widget is None
